import json
import logging
import pprint

from gallium.interface import ICommand

from .core             import Core
from .drivers.rabbitmq import Driver, NoConnectionError
from .observer         import Observer, SYNC_START
from .helper           import prepare_logger


class Server(ICommand):
    """ Run the sample """
    def identifier(self):
        return 'sample.observe'

    def define(self, parser):
        parser.add_argument(
            '--debug',
            '-d',
            action = 'store_true'
        )

        parser.add_argument(
            '--bind-url',
            '-b',
            default='amqp://guest:guest@172.17.0.1:5672/%2F'
        )

    def execute(self, args):
        # logging.basicConfig(level = logging.DEBUG if args.debug else logging.INFO)
        prepare_logger(logging.DEBUG if args.debug else logging.INFO)

        driver  = Driver(args.bind_url)
        service = Observer(driver)

        # In this example, delegation is disabled.
        # vireo.open('vireo.sample.primary', delegation_ttl = 5000)
        # vireo.on('vireo.sample.primary.delegated', lambda x: print('vireo.sample.primary.delegated: {}'.format(x)))

        def wrapper(label, data):
            print('{}:'.format(label))
            pprint.pprint(data, indent = 2)

        service.on('vireo.sample.direct',           lambda x: wrapper('vireo.sample.direct',    x))
        service.on('vireo.sample.secondary',        lambda x: wrapper('vireo.sample.secondary', x))
        service.on('vireo.sample.direct.resumable', lambda x: wrapper('vireo.sample.direct',    x), resumable = True)

        service.on_broadcast('vireo.sample.broadcast.one', lambda x: wrapper('vireo.sample.broadcast.one', x))
        service.on_broadcast('vireo.sample.broadcast.two', lambda x: wrapper('vireo.sample.broadcast.two', x))

        service.join(SYNC_START)


class EventEmitter(ICommand):
    """ Emit an event """
    def identifier(self):
        return 'event.emit'

    def define(self, parser):
        parser.add_argument(
            '--debug',
            '-d',
            action = 'store_true'
        )

        parser.add_argument(
            'event_name',
            help = 'The name of the event (e.g., "sample.primary")'
        )

        parser.add_argument(
            'event_data',
            help  = 'The JSON-compatible string data of the event',
            nargs = '?'
        )

        parser.add_argument(
            '--bind-url',
            '-b',
            default='amqp://guest:guest@172.17.0.1:5672/%2F'
        )

    def execute(self, args):
        prepare_logger(logging.DEBUG if args.debug else logging.INFO)

        driver  = Driver(args.bind_url)
        service = Core(driver)

        service.emit(args.event_name, json.loads(args.event_data) if args.event_data else None)

class EventBroadcaster(ICommand):
    """ Broadcast an event """
    def identifier(self):
        return 'event.broadcast'

    def define(self, parser):
        parser.add_argument(
            '--debug',
            '-d',
            action = 'store_true'
        )

        parser.add_argument(
            'event_name',
            help = 'The name of the event (e.g., "sample.primary")'
        )

        parser.add_argument(
            'event_data',
            help  = 'The JSON-compatible string data of the event',
            nargs = '?'
        )

        parser.add_argument(
            '--bind-url',
            '-b',
            default='amqp://guest:guest@172.17.0.1:5672/%2F'
        )

    def execute(self, args):
        prepare_logger(logging.DEBUG if args.debug else logging.INFO)

        driver  = Driver(args.bind_url)
        service = Core(driver)

        service.broadcast(args.event_name, json.loads(args.event_data) if args.event_data else None)



import logging

_logger      = None
_log_handler = None


def log(level, *args, **kwargs):
    global _logger

    if not _logger:
        prepare__logger(logging.INFO)

    getattr(_logger, level)(*args, **kwargs)


def prepare_logger(level):
    global _logger

    formatter = logging.Formatter('{name} @ {asctime} [{levelname}] {message}', style='{')

    _logger = logging.getLogger('vireo')
    _logger.setLevel(level)

    log_handler = logging.StreamHandler()
    log_handler.setLevel(level)
    log_handler.setFormatter(formatter)

    _logger.addHandler(log_handler)

    return _logger

def fill_in_the_blank(main_dict, default_dict):
    for default_key, default_value in default_dict.items():
        if default_key in main_dict:
            continue

        main_dict[default_key] = default_value

    return main_dict



import json
import threading
import time

from pika            import BasicProperties
from pika.exceptions import ConnectionClosed, ChannelClosed

from ...helper import fill_in_the_blank, log

from .consumer  import Consumer
from .exception import NoConnectionError, SubscriptionNotAllowedError
from .helper    import active_connection, SHARED_TOPIC_EXCHANGE_NAME, SHARED_SIGNAL_CONNECTION_LOSS


class Driver(object):
    def __init__(self, url):
        self._url             = url
        self._async_listener  = None
        self._shared_stream   = []
        self._consumers       = []
        self._has_term_signal = False
        self._active_routes   = []

    def setup_async_cleanup(self):
        """ Synchronously join all consumers."""
        if self._async_listener and self._async_listener.is_alive():
            raise SubscriptionNotAllowedError('Unable to consume messages as this driver is currently active.')

        self._async_listener = threading.Thread(target = self.join)
        self._async_listener.start()

    def stop_consuming(self):
        self._has_term_signal = True

    def join(self):
        """ Synchronously join all consumers."""
        try:
            while True:
                if self._has_term_signal:
                    log('info', 'Stopping all route listeners')

                    break

                if SHARED_SIGNAL_CONNECTION_LOSS in self._shared_stream:
                    log('error', 'Unexpected connection loss detected')
                    log('warning', 'Terminating all route listeners')

                    break

                time.sleep(1)
        except KeyboardInterrupt:
            log('warning', 'SIGTERM received')
            log('debug', 'Terminating all route listeners')

        connection_losed = SHARED_SIGNAL_CONNECTION_LOSS in self._shared_stream

        for consumer in self._consumers:
            if not connection_losed:
                if not consumer.is_alive():
                    log('info', 'Route {}: Already stopped listening (not alive).'.format(consumer.route))

                    continue

                log('debug', 'Route {}: Sending the signal to stop listening.'.format(consumer.route))
                consumer.stop()

            try:
                log('debug', 'Route {}: Terminating the listener.'.format(consumer.route))
                consumer._stop()
            except AssertionError: # this is raised if the thread lock is still locked.
                log('warning', 'Route {}: Failed to stop cleanly'.format(consumer.route))

            if not consumer.is_alive():
                log('info', 'Route {}: Termination confirmed (killed)'.format(route_listener.route))

                continue

            log('debug', 'Route {}: Waiting the listener to join back to the parent thread.'.format(consumer.route))
            consumer.join()
            log('info', 'Route {}: Termination confirmed (joined).'.format(consumer.route))

        if connection_losed:
            raise NoConnectionError('Unexpectedly losed the connection during message consumption')

    def publish(self, route, message, options = None):
        """ Synchronously publish a message

            :param str route:   the route
            :param str message: the message
        """
        default_parameters = {
            'exchange'    : '',
            'routing_key' : route or '',
            'body'        : json.dumps(message),
            'properties'  : BasicProperties(content_type = 'application/json'),
        }

        options = fill_in_the_blank(options or {}, default_parameters)

        with active_connection(self._url) as channel:
            try:
                log('debug', 'Declaring: route={}'.format(route))
                channel.queue_declare(queue = route, passive = True)
                log('debug', 'Declared: route={}'.format(route))

                log('debug', 'Publishing: route={} message={} options={}'.format(route, message, options))
                channel.basic_publish(**options)
                log('debug', 'Published: route={} message={} options={}'.format(route, message, options))
            except ConnectionClosed:
                raise NoConnectionError('Unexpectedly losed the connection while publishing a message')

    def declare_queue_with_delegation(self, origin_queue_name, ttl, fallback_queue_name = None,
                                      common_queue_options = None, exchange_options = None):

        actual_fallback_queue_name = fallback_queue_name or '{}.delegated'.format(origin_queue_name)
        exchange_name              = 'fallback/{}/{}'.format(origin_queue_name, actual_fallback_queue_name)

        exchange_options     = fill_in_the_blank(exchange_options or {}, {'internal': True})
        common_queue_options = common_queue_options or {}

        default_fallback_queue_options = {'auto_delete': False}
        default_origin_queue_options   = {
            'auto_delete' : False,
            'arguments'   : {
                'x-dead-letter-exchange'    : exchange_name,
                'x-dead-letter-routing-key' : actual_fallback_queue_name,
                'x-message-ttl'             : ttl,
            }
        }

        fallback_queue_options = fill_in_the_blank(dict(common_queue_options), default_fallback_queue_options)
        origin_queue_options   = fill_in_the_blank(dict(common_queue_options), default_origin_queue_options)

        fill_in_the_blank(exchange_options, {'exchange': exchange_name, 'exchange_type': 'direct'})

        with active_connection(self._url) as channel:
            try:
                channel.exchange_declare(**exchange_options)

                self.declare_queue(actual_fallback_queue_name, fallback_queue_options)

                channel.queue_bind(
                    queue    = actual_fallback_queue_name,
                    exchange = exchange_name,
                )

                self.declare_queue(origin_queue_name, origin_queue_options)
            except ConnectionClosed:
                raise NoConnectionError('Unexpectedly losed the connection while orchestrating queues and exchange for delegation')

    def broadcast(self, route, message, options = None):
        default_parameters = {
            'exchange'    : '',
            'routing_key' : route or '',
            'body'        : json.dumps(message),
            'properties'  : BasicProperties(content_type = 'application/json'),
        }

        options = fill_in_the_blank(options or {}, default_parameters)

        if 'exchange' not in options or not options['exchange']:
            options['exchange'] = SHARED_TOPIC_EXCHANGE_NAME

        exchange_name = options['exchange']

        with active_connection(self._url) as channel:
            try:
                log('debug', 'Declaring a shared topic exchange')

                channel.exchange_declare(
                    exchange      = exchange_name,
                    exchange_type = 'topic',
                    passive       = False,
                    durable       = True,
                    auto_delete   = False,
                )

                log('debug', 'Declared a shared topic exchange')

                log('debug', 'Broadcasting: route={} message={} options={}'.format(route, message, options))
                channel.basic_publish(**options)
                log('debug', 'Broadcasted: route={} message={} options={}'.format(route, message, options))

            except ConnectionClosed:
                raise NoConnectionError('Unexpectedly losed the connection while broadcasting an event')

    def observe(self, route, callback, resumable, distributed, options = None):
        consumer = Consumer(self._url, route, callback, self._shared_stream, resumable, distributed, options)
        consumer.start()

        self._consumers.append(consumer)



import contextlib

from pika            import BlockingConnection
from pika.connection import URLParameters
from pika.exceptions import ConnectionClosed, ChannelClosed

from ...helper  import fill_in_the_blank, log
from .exception import NoConnectionError

SHARED_SIGNAL_CONNECTION_LOSS = 1
SHARED_TOPIC_EXCHANGE_NAME    = 'vireo_default_topic_r0'


def get_blocking_queue_connection(url):
    init_params = URLParameters(url)

    return BlockingConnection(init_params)


@contextlib.contextmanager
def active_connection(url):
    log('debug', 'Connecting')

    try:
        connection = get_blocking_queue_connection(url)
        channel    = connection.channel()
    except ChannelClosed:
        raise NoConnectionError('Failed to communicate while opening an active channel')
    except ConnectionClosed:
        raise NoConnectionError('Failed to connect while opening an active connection')

    log('debug', 'Connected and channel opened')

    log('debug', 'Yielding the opened channel')

    yield channel

    log('debug', 'Regained the opened channel')

    log('debug', 'Disconnecting')

    try:
        channel.close()
        connection.close()

        log('debug', 'Disconnected')
    except ChannelClosed:
        log('warning', 'Already closed the channel') # bypassed if the connection is no longer available.
    except ConnectionClosed:
        log('warning', 'Already disconnected') # bypassed if the connection is no longer available.



import json
import threading

from pika.exceptions import ConnectionClosed, ChannelClosed

from ...helper import log

from .helper import active_connection, fill_in_the_blank, SHARED_TOPIC_EXCHANGE_NAME


class Consumer(threading.Thread):
    def __init__(self, url, route, callback, shared_stream, resumable, distributed, queue_options):
        super().__init__(daemon = True)

        self.url            = url
        self.route          = route
        self.callback       = callback
        self.resumable      = resumable
        self.distributed    = distributed
        self.queue_options  = queue_options
        self._shared_stream = shared_stream
        self._channel       = None
        self._queue_name    = None

    def run(self):
        with active_connection(self.url) as channel:
            self._channel = channel

            self._queue_name = self._declare_topic_queue(channel) if self.distributed else self._declare_shared_queue(channel)

            # Declare the callback wrapper for this route.
            def callback_wrapper(channel, method_frame, header_frame, body):
                log('debug', 'Method Frame: {}'.format(method_frame))
                log('debug', 'Header Frame: {}'.format(header_frame))
                log('debug', 'Body: {}'.format(header_frame))

                self.callback(json.loads(body.decode('utf8')))

                channel.basic_ack(delivery_tag = method_frame.delivery_tag)

            log('debug', 'Listening to {}'.format(self._debug_route_name()))

            channel.basic_consume(callback_wrapper, self._queue_name)

            # NOTE there is a bug in start_consuming that prevents stop_consuming from cleanly
            #      stopping message consumption. The following is a hack suggested in StackOverflow.
            # channel.start_consuming()
            try:
                while channel._consumer_infos:
                    channel.connection.process_data_events(time_limit = 1)
            except ConnectionClosed:
                log('warning', 'Unexpected connection loss while listening to {}'.format(self._debug_route_name()))

                self._shared_stream.append(SHARED_SIGNAL_CONNECTION_LOSS)

            log('debug', 'Stopped listening to {}'.format(self._debug_route_name()))

    def stop(self):
        log('debug', 'Stopping listening to {}'.format(self._debug_route_name()))
        self._channel.stop_consuming()

    def _debug_route_name(self):
        return '{} ({})'.format(self.route, self._queue_name)

    def _declare_shared_queue(self, channel):
        queue_options = fill_in_the_blank(
            {
                'auto_delete': not self.resumable,
                'queue'      : self.route,
            },
            self.queue_options or {}
        )

        channel.queue_declare(**queue_options)

        log('info', 'Declared a queue "{}"'.format(self.route))

        return self.route

    def _declare_topic_queue(self, channel):
        # Currently not supporting resumability.
        queue_options = fill_in_the_blank(
            {
                'auto_delete': True,
                'queue'      : '',
            },
            self.queue_options or {}
        )

        response        = channel.queue_declare(**queue_options)
        temp_queue_name = response.method.queue

        log('info', 'Declared a temporary queue "{}"'.format(temp_queue_name))

        log('info', 'Binding a temporary queue "{}" to route {}'.format(temp_queue_name, self.route))
        channel.queue_bind(temp_queue_name, SHARED_TOPIC_EXCHANGE_NAME, self.route)
        log('info', 'Bound a temporary queue "{}" to route {}'.format(temp_queue_name, self.route))

        return temp_queue_name



from .core   import Core
from .helper import log

ASYNC_START = 1
SYNC_START  = 2


class UnknownRunningModeError(RuntimeError):
    """ Error for unknown running mode """


class Observer(Core):
    """ Event Observer """
    # def open(self, event_name, options = None, delegation_ttl = None):
    #     """ Prepare to observe an event.
    #
    #         :param str event_name:     the name of the queue
    #         :param str options:        the option for the driver
    #         :param int delegation_ttl: the TTL for the delegation*
    #
    #         To enable delegation, ``delegation_ttl`` must not be set to ``None`` or ``0`` or ``False``.
    #
    #         The delegation of the event happens when there exists no listener to that event. Then,
    #         the message will be transfered to the delegated event, which is similar to the given
    #         event name but suffixed with ``.delegated``. For instance, given an event name
    #         "outbound_mail_delivery", the delegated event name will be "outbound_mail_delivery.delegated".
    #
    #         To handle the delegated event, simply listen to the delegated event. For example,
    #         continuing from the previous example, you can write ``on('outbound_mail_delivery.delegated', lambda x: foo(x))``.
    #
    #         For example,
    #
    #         .. code-block:: python
    #
    #             app.open('foo')
    #
    #         .. note:: the delegation might change in the future to simplify the interface.
    #     """
    #     if delegation_ttl:
    #         log('debug', 'Preparing to observe event "{}" with delegation after {} seconds'.format(event_name, delegation_ttl / 1000))
    #
    #         self._driver.declare_queue_with_delegation(
    #             event_name,
    #             delegation_ttl,
    #             common_queue_options = options
    #         )
    #
    #         log('debug', 'Ready to observe event "{}" with delegation after {} seconds'.format(event_name, delegation_ttl / 1000))
    #
    #         return
    #
    #     log('debug', 'Preparing to observe event "{}"'.format(event_name))
    #
    #     self._driver.declare_queue(event_name, options or {})
    #
    #     log('debug', 'Ready to observe event "{}"'.format(event_name))
    #
    # def close(self, event_name, options = None):
    #     """ Clean up after observing an event.
    #
    #         :param str event_name: the name of the queue
    #         :param str options:    the option for the driver
    #     """
    #     log('debug', 'Preparing to stop observing event "{}"'.format(event_name))
    #
    #     self._driver.delete_queue(event_name, options or {})
    #
    #     log('debug', 'Event "{}" may not be observed properly at this point.'.format(event_name))

    def on(self, event_name, callback, resumable = False):
        """ Listen to an event with a callback function.

            :param str event_name: the name of the event
            :param callable callback: the callback callable
            :param bool resumable: the flag to indicate whether the event consumption can be resumed (as the data stream will never be deleted).

            The callback is a callable object, e.g., function, class method, lambda object, which
            takes only one parameter which is a JSON-decoded object.

            For example,

            .. code-block:: python

                def on_foo(self, message):
                    print('on_foo:', message)

                app.on('foo', on_foo)
                app.on('foo.lambda', lambda x: print('foo_lambda:', x))
        """
        self._driver.observe(event_name, callback, resumable, False)

    def on_broadcast(self, event_name, callback):
        """ Listen to an distributed event with a callback function.

            The callback is a callable object, e.g., function, class method, lambda object, which
            takes only one parameter which is a JSON-decoded object.

            For example,

            .. code-block:: python

                def on_foo(self, message):
                    print('on_foo:', message)

                app.on('foo', on_foo)
                app.on('foo.lambda', lambda x: print('foo_lambda:', x))
        """
        self._driver.observe(event_name, callback, False, True)

    def join(self, running_mode = SYNC_START):
        """ Start the observation

            There are two mode: synchronous (``vireo.observer.SYNC_START``) and asynchronous
            (``vireo.observer.ASYNC_START``) starts.

            .. code-block:: python

                app.start(ASYNC_START)
        """
        if running_mode == ASYNC_START:
            self._driver.setup_async_cleanup()

            return

        if running_mode == SYNC_START:
            self._driver.join()

            return

        raise UnknownRunningModeError(running_mode)

    def stop(self):
        """ Stop the observation

            .. code-block:: python

                app.stop()
        """
        self._driver.stop_consuming()
